# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import GridUtils
import PGUtils
from Game import GameState

def enumerate_body_neighbors(grid, positions, val, empty_only=False):
    """ Returns orthogonal neighbors of a "body" (connected group of pieces) as (position, value) tuples.
        If empty_only is true, only returns empty positions (where value is 0) """
    seen = set()
    for p in positions:
        for np,nval in GridUtils.enumerate_ortho_neighbors(grid, p):
            if nval!=val and (not empty_only or nval==0):
                if np not in seen: yield (np, nval)
                seen.add(np)
                
def has_freedom(grid, positions, val):
    """ Returns true if any position in the group borders (orthogonally) an empty cell """
    return any(enumerate_body_neighbors(grid, positions, val, True))
                
class SymbioGameState(GameState):
    def successors(self):
        pnum = self.current_player
        bodies = GridUtils.connected_groups(self.grid, pnum)
        opp_bodies = {}
        # each piece in the body can move to any of the body's neighbors
        for body in bodies:
            if not has_freedom(self.grid, body, pnum): continue
            for np,nval in enumerate_body_neighbors(self.grid, body, pnum):
                # valid if neighbor is empty, or enemy body without freedom
                good = (nval==0)
                if not good:
                    # find the opponent's body containing the destination
                    if nval not in opp_bodies: 
                        opp_bodies[nval] = list(GridUtils.connected_groups(self.grid, nval))
                    opp_body = PGUtils.first_item(b for b in opp_bodies[nval] if np in b)
                    good = not has_freedom(self.grid, opp_body, nval)
                if good:
                    for bp in body:
                        # we have a valid move, get all drop positions which border a resulting body
                        nextgrid = self.grid.copy()
                        nextgrid[bp] = 0
                        nextgrid[np] = pnum
                        for droppos in GridUtils.enumerate_coords_with_value(nextgrid, 0):
                            if any(dpval==pnum for _,dpval in GridUtils.enumerate_ortho_neighbors(nextgrid, droppos)):
                                ng2 = nextgrid.copy()
                                ng2[droppos] = pnum
                                yield ((bp,np,droppos), self.next_state_with_grid(ng2))
                        
    def has_scores(self):
        return False

    def animations_for_move(self, move, newstate):
        import PGPieceAnimation
        animations = [PGPieceAnimation.move_animation(move[0], move[1])]
        # remove opposing piece if at destination
        if self.grid[move[1]]!=0:
            animations.append(PGPieceAnimation.remove_animation(move[1]))
        # add drop animation
        animations.append(PGPieceAnimation.new_piece_animation(move[2], self.current_player, animations[0].end_time()))
        return PGPieceAnimation.AnimationGroup(animations)
        
        
from GenericAI import GenericAI

class SymbioAI(GenericAI):
    def game_utility(self, game):
        utility = 0
        for pnum in xrange(1, 1+game.num_players):
            bodies = GridUtils.connected_groups(game.grid, pnum)
            for body in bodies:
                # free cells increase in value as sqrt(N), having zero liberties is bad
                freedom = sum(1 for _ in enumerate_body_neighbors(game.grid, body, pnum, True))
                fval = 100*(freedom**0.5) if freedom>0 else -2
                utility += (len(body)*fval * (1 if game.current_player==pnum else -1))
        return utility
        